@@ -9,7 +9,9 @@ import datetime
 import os
 import re
 import time
+from collections.abc import Awaitable
 from typing import Any
+from typing import Callable
 from typing import Optional
 from urllib.parse import urlencode
 
@@ -22,6 +24,7 @@ from speechmatics.rt import TranscriptionConfig
 
 from . import __version__
 from ._logging import get_logger
+from ._models import AgentClientMessageType
 from ._models import AgentServerMessageType
 from ._models import AnnotationFlags
 from ._models import AnnotationResult
@@ -37,7 +40,6 @@ from ._models import SpeechFragment
 from ._models import VoiceAgentConfig
 
 DEBUG_MORE = os.getenv("SPEECHMATICS_DEBUG_MORE", "0").lower() in ["1", "true"]
-PREVIEW_FEATURES = os.getenv("SPEECHMATICS_PREVIEW_FEATURES", "0").lower() in ["1", "true"]
 
 
 if DEBUG_MORE:
@@ -95,6 +97,10 @@ class VoiceAgentClient(AsyncClient):
             language_pack_info=LanguagePackInfo.from_dict({}),
         )
 
+        # STT message received queue
+        self._stt_message_queue: asyncio.Queue[Callable[[], Awaitable[None]]] = asyncio.Queue()
+        self._stt_queue_task: Optional[asyncio.Task] = None
+
         # Timing info
         self._total_time: float = 0
         self._total_bytes: int = 0
@@ -126,20 +132,12 @@ class VoiceAgentClient(AsyncClient):
 
         # Timers for EndOfUtterance and EndOfTurn
         self._finalize_task: Optional[asyncio.Task] = None
-        self._eot_task: Optional[asyncio.Task] = None
-
-        # Segment processor
-        self._processor_wait_time: float = 0.005
-        self._processor_task: Optional[asyncio.Task] = None
-
-        # Emitter task
-        self._finals_emitter_task: Optional[asyncio.Task] = None
 
         # Metrics emitter task
         self._metrics_emitter_interval: float = 10.0
         self._metrics_emitter_task: Optional[asyncio.Task] = None
 
-        # Audi sampling info
+        # Audio sampling info
         self._audio_sample_rate: float = self._audio_format.sample_rate * 1.0
         self._audio_sample_width: float = {
             AudioEncoding.PCM_F32LE: 4.0,
@@ -235,6 +233,8 @@ class VoiceAgentClient(AsyncClient):
         # Recognition started event
         @self.once(ServerMessageType.RECOGNITION_STARTED)
         def _evt_on_recognition_started(message: dict[str, Any]) -> None:
+            if DEBUG_MORE:
+                self._logger.debug(json.dumps(message))
             self._is_ready_for_audio = True
             self._session = SessionInfo(
                 config=self._config,
@@ -246,22 +246,35 @@ class VoiceAgentClient(AsyncClient):
         # Partial transcript event
         @self.on(ServerMessageType.ADD_PARTIAL_TRANSCRIPT)
         def _evt_on_partial_transcript(message: dict[str, Any]) -> None:
-            self._handle_transcript(message, is_final=False)
+            if DEBUG_MORE:
+                self._logger.debug(json.dumps(message))
+
+            async def _handle() -> None:
+                await self._handle_transcript(message, is_final=False)
+
+            self._stt_message_queue.put_nowait(_handle)
 
         # Final transcript event
         @self.on(ServerMessageType.ADD_TRANSCRIPT)
         def _evt_on_final_transcript(message: dict[str, Any]) -> None:
-            self._handle_transcript(message, is_final=True)
+            if DEBUG_MORE:
+                self._logger.debug(json.dumps(message))
+
+            async def _handle() -> None:
+                await self._handle_transcript(message, is_final=True)
+
+            self._stt_message_queue.put_nowait(_handle)
 
         # End of utterance event
-        if self._end_of_utterance_mode == EndOfUtteranceMode.FIXED:
+        @self.on(ServerMessageType.END_OF_UTTERANCE)
+        def _evt_on_end_of_utterance(message: dict[str, Any]) -> None:
+            if DEBUG_MORE:
+                self._logger.debug(json.dumps(message))
 
-            @self.on(ServerMessageType.END_OF_UTTERANCE)
-            def _evt_on_end_of_utterance(message: dict[str, Any]) -> None:
-                self.emit(
-                    AgentServerMessageType.END_OF_TURN,
-                    {"message": AgentServerMessageType.END_OF_TURN.value, "metadata": message.get("metadata", {})},
-                )
+            async def _handle() -> None:
+                await self._emit_segments(finalize=True, end_of_turn=True)
+
+            self._stt_message_queue.put_nowait(_handle)
 
     async def connect(self) -> None:
         """Connect to the Speechmatics API.
@@ -279,6 +292,9 @@ class VoiceAgentClient(AsyncClient):
             )
             return
 
+        # Start the processor task
+        self._stt_queue_task = asyncio.create_task(self._run_stt_queue())
+
         # Connect to API
         try:
             await self.start_session(
@@ -312,6 +328,11 @@ class VoiceAgentClient(AsyncClient):
             self._is_ready_for_audio = False
             self._stop_metrics_task()
 
+        # Stop the task
+        if self._stt_queue_task:
+            self._stt_queue_task.cancel()
+            self._stt_queue_task = None
+
     def update_diarization_config(self, config: DiarizationSpeakerConfig) -> None:
         """Update the diarization configuration.
 
@@ -324,6 +345,33 @@ class VoiceAgentClient(AsyncClient):
         """
         self._dz_config = config
 
+    def _start_stt_queue(self) -> None:
+        """Start the STT message queue."""
+        self._stt_queue_task = asyncio.create_task(self._run_stt_queue())
+
+    async def _run_stt_queue(self) -> None:
+        """Run the STT message queue."""
+        while True:
+            callback = await self._stt_message_queue.get()
+
+            try:
+                if asyncio.iscoroutine(callback):
+                    await callback
+                elif asyncio.iscoroutinefunction(callback):
+                    await callback()
+                elif callable(callback):
+                    result = callback()
+                    if asyncio.iscoroutine(result):
+                        await result
+
+            except Exception:
+                self._logger.warning("Exception in STT message queue", exc_info=True)
+
+    def _stop_stt_queue(self) -> None:
+        """Stop the STT message queue."""
+        if self._stt_queue_task:
+            self._stt_queue_task.cancel()
+
     async def send_audio(self, payload: bytes) -> None:
         """Send an audio frame through the WebSocket.
 
@@ -383,50 +431,29 @@ class VoiceAgentClient(AsyncClient):
             self._metrics_emitter_task.cancel()
             self._metrics_emitter_task = None
 
-    def _handle_transcript(self, message: dict[str, Any], is_final: bool) -> None:
-        """Handle the partial and final transcript events.
-
-        Args:
-            message: The new Partial or Final from the STT engine.
-            is_final: Whether the data is final or partial.
-        """
-        # Handle async
-        asyncio.create_task(self._handle_transcript_async(message, is_final))
-
-    async def _handle_transcript_async(self, message: dict[str, Any], is_final: bool) -> None:
+    async def _handle_transcript(self, message: dict[str, Any], is_final: bool) -> None:
         """Handle the partial and final transcript events (async).
 
+        As `AddTranscript` messages are _always_ followed by `AddPartialTranscript` messages,
+        we can skip processing. Also skop if there are no fragments in the buffer.
+
         Args:
             message: The new Partial or Final from the STT engine.
             is_final: Whether the data is final or partial.
         """
 
-        # Debug
-        if DEBUG_MORE:
-            self._logger.debug(json.dumps(message))
-
         # Add the speech fragments
         fragments_available = await self._add_speech_fragments(
             message=message,
             is_final=is_final,
         )
 
-        # Skip if unchanged
-        if not fragments_available:
+        # Skip if no fragments or `AddTranscript`
+        if not fragments_available or is_final:
             return
 
-        # Clear any existing timer
-        if self._processor_task is not None:
-            self._processor_task.cancel()
-
-        # Send transcription frames after delay
-        async def process_after_delay(delay: float) -> None:
-            await asyncio.sleep(delay)
-            await self._process_speech_fragments()
-            self._processor_task = None
-
-        # Send frames after delay
-        self._processor_task = asyncio.create_task(process_after_delay(self._processor_wait_time))
+        # Process
+        await self._process_speech_fragments()
 
     def _calculate_ttfb(self, end_time: float) -> None:
         """Calculate the time to first text.
@@ -558,6 +585,14 @@ class VoiceAgentClient(AsyncClient):
             self._speech_fragments.extend(fragments)
             self._speech_fragments.sort(key=lambda x: x.idx)
 
+            # Remove fragment at head that is for previous
+            if (
+                self._speech_fragments
+                and self._speech_fragments[0].is_punctuation
+                and self._speech_fragments[0].attaches_to == "previous"
+            ):
+                self._speech_fragments.pop(0)
+
             # Debug the fragments
             if DEBUG_MORE:
                 debug_payload = {
@@ -604,13 +639,6 @@ class VoiceAgentClient(AsyncClient):
             # Trim the view
             self._current_view.trim(start_time=self._current_view.start_time, end_time=last_segment.end_time)
 
-        # Emit the segments
-        asyncio.create_task(self._emit_segments())
-
-        # Only do this for ADAPTIVE end of utterance
-        if self._end_of_utterance_mode == EndOfUtteranceMode.ADAPTIVE:
-            """If ADAPTIVE, then calculate whether the turn has completed."""
-
             # Compare previous view to this view
             if self._previous_view:
                 changes = FragmentUtils.compare_views(self._session, self._previous_view, self._current_view)
@@ -620,6 +648,19 @@ class VoiceAgentClient(AsyncClient):
             # Update the previous view
             self._previous_view = self._current_view
 
+        # Catch no changes for ADAPTIVE mode
+        if self._end_of_utterance_mode == EndOfUtteranceMode.ADAPTIVE and not changes.any(
+            AnnotationFlags.NEW, AnnotationFlags.UPDATED_FULL
+        ):
+            return
+
+        # Emit the segments
+        await self._emit_segments()
+
+        # Only do this for ADAPTIVE end of utterance
+        if self._end_of_utterance_mode == EndOfUtteranceMode.ADAPTIVE:
+            """If ADAPTIVE, then calculate whether the turn has completed."""
+
             # Calculate delay to end of utterance
             should_emit, emit_final_delay = self._calc_delay_to_end_of_utterance(self._current_view, changes)
 
@@ -747,14 +788,20 @@ class VoiceAgentClient(AsyncClient):
         """
 
         async def emit() -> None:
+            # Yield until the delay has passed
             if ttl is not None and ttl > 0:
                 await asyncio.sleep(ttl)
-            if PREVIEW_FEATURES:
-                await self.send_message({"message": "Finalize"})
-            else:
-                await self._emit_segments(finalize=True)
 
-        asyncio.create_task(emit())
+            # Finalize the segments using `Finalize` message to STT
+            if self._config.enable_preview_features:
+                await self.send_message({"message": AgentClientMessageType.FINALIZE_TURN})
+
+            # Force finalizing of segments in the buffer
+            else:
+                await self._emit_segments(finalize=True, end_of_turn=True)
+
+        # Add to queue
+        self._stt_message_queue.put_nowait(emit)
 
     async def _emit_segments(self, finalize: bool = False, end_of_turn: bool = False) -> None:
         """Emit segments to listeners.
@@ -768,17 +815,12 @@ class VoiceAgentClient(AsyncClient):
             end_of_turn: Whether to emit an end of turn event.
         """
 
-        # Metadata
-        if self._current_view:
-            metadata = {"start_time": self._current_view.start_time, "end_time": self._current_view.end_time}
-        elif self._previous_view:
-            metadata = {"start_time": self._previous_view.start_time, "end_time": self._previous_view.end_time}
-        else:
-            metadata = None
-
         # Lock the speech fragments
         if self._current_view and self._current_view.segment_count > 0:
             async with self._speech_fragments_lock:
+                # Metadata
+                metadata = {"start_time": self._current_view.start_time, "end_time": self._current_view.end_time}
+
                 # Force finalize
                 if finalize:
                     final_segments = self._current_view.segments
@@ -823,7 +865,11 @@ class VoiceAgentClient(AsyncClient):
                 self._update_current_view()
 
         # Emit end of turn
-        if end_of_turn:
+        if end_of_turn and self._previous_view:
+            # Metadata (for LAST view)
+            metadata = {"start_time": self._previous_view.start_time, "end_time": self._previous_view.end_time}
+
+            # Emit
             self.emit(
                 AgentServerMessageType.END_OF_TURN,
                 {
@@ -832,6 +878,9 @@ class VoiceAgentClient(AsyncClient):
                 },
             )
 
+            # Reset the previous view
+            self._previous_view = None
+
     def _vad_evaluation(self, fragments: list[SpeechFragment]) -> None:
         """Emit a VAD event.
 
