@@ -174,6 +174,9 @@ class VoiceAgentConfig:
         include_results: Include word data in the response. This is useful for debugging and
             understanding the STT engine's behavior. Defaults to False.
 
+        enable_preview_features: Enable preview features using a preview endpoint provided by
+            Speechmatics. Defaults to False.
+
         sample_rate: Audio sample rate for streaming. Defaults to 16000.
         audio_encoding: Audio encoding format. Defaults to AudioEncoding.PCM_S16LE.
     """
@@ -202,6 +205,7 @@ class VoiceAgentConfig:
 
     # Advanced features
     include_results: bool = False
+    enable_preview_features: bool = False
 
     # Audio
     sample_rate: int = 16000
@@ -234,7 +238,7 @@ class AgentClientMessageType(str, Enum):
         >>> await client.send_message(end_message)
     """
 
-    FINALIZE_TURN = "FinalizeTurn"
+    FINALIZE_TURN = "Finalize"
     END_OF_STREAM = "EndOfStream"
     GET_SPEAKERS = "GetSpeakers"
 
@@ -326,6 +330,7 @@ class AnnotationFlags(str, Enum):
     UPDATED_FINALS = "updated_finals"
     UPDATED_PARTIALS = "updated_partials"
     UPDATED_SPEAKERS = "updated_speakers"
+    FINALIZED = "finalized"
 
     # Content of segments
     ONLY_ACTIVE_SPEAKERS = "only_active_speakers"
@@ -528,12 +533,12 @@ class SpeakerSegment:
     @property
     def start_time(self) -> float:
         """Return the start time of the segment."""
-        return self.fragments[0].start_time
+        return self.fragments[0].start_time if self.fragments else 0.0
 
     @property
     def end_time(self) -> float:
         """Return the end time of the segment."""
-        return self.fragments[-1].end_time
+        return self.fragments[-1].end_time if self.fragments else 0.0
 
     def to_dict(self, include_results: bool = False) -> dict[str, Any]:
         """Return a dictionary representation of the object."""
@@ -543,6 +548,7 @@ class SpeakerSegment:
             "timestamp": self.timestamp,
             "language": self.language,
             "text": self.text,
+            "annotation": self.annotation,
             **({"results": [f.result for f in self.fragments]} if include_results else {}),
         }
 
@@ -587,11 +593,11 @@ class SpeakerSegmentView:
 
     @property
     def start_time(self) -> float:
-        return self.fragments[0].start_time
+        return self.fragments[0].start_time if self.fragments else 0.0
 
     @property
     def end_time(self) -> float:
-        return self.fragments[-1].end_time
+        return self.fragments[-1].end_time if self.fragments else 0.0
 
     @property
     def final_count(self) -> int:
@@ -923,6 +929,8 @@ class FragmentUtils:
     ) -> AnnotationResult:
         """Compare two SpeakerSegmentView objects and return the differences.
 
+        View 1 (new) is compared to view 2 (old).
+
         Args:
             session: SessionInfo object.
             view1: The first SpeakerSegmentView object to compare.
@@ -964,5 +972,9 @@ class FragmentUtils:
         elif view1.segment_count > 0:
             result.add(AnnotationFlags.NEW)
 
+        # Finalized (last segment only has finals)
+        if view1.segment_count > 0 and view1.partial_count == 0:
+            result.add(AnnotationFlags.FINALIZED)
+
         # Return the result
         return result
